

from NikeCA import NikeCA, Snowflake, QA
from NikeSF import Snowflake
from NikeQA import QA
from _SnowflakeDependencies import SnowflakeDependencies
from _IMPSummaryDashboard import IMPSummaryDashboard
from _BuildSearchQuery import BuildSearchQuery




import NikeCA
import _QA
import pandas as pd
import _GitHub


class QA(_QA.QA, _GitHub.GitHub):
    from NikeSF import Snowflake

    """
    A class for interacting with the data and doing basic QA
    """

    def __init__(self, df: pd.DataFrame = None, df2: pd.DataFrame = None, ds1_nm: str = 'Source #1',
                 ds2_nm: str = 'Source #2', case_sens: bool = True, print_analysis: bool = True, check_match_by=None,
                 breakdown_grain=None, token: str = None, tables: list | str = None, save_path: str = None,
                 username: str | None = None, warehouse: str | None = None, role: str | None = None,
                 database: str | None = None, schema: str | None = None, git_org: str | None = None):

        """

        :param df:
        :param df2: DataFrame | None = None
        :param ds1_nm: str = 'Source #1'
        :param ds2_nm: str = 'Source #2'
        :param case_sens: bool = True
        :param print_analysis: bool = True
        :param check_match_by: Any = None
        :param breakdown_grain: Any = None
        :param token: str = None

        """
        self.__git_org = git_org
        self.__df = df
        self.__df2 = df2
        self.__ds1_nm = ds1_nm
        self.__ds2_nm = ds2_nm
        self.__case_sens = case_sens
        self.__print_analysis = print_analysis
        self.__check_match_by = check_match_by
        self.__breakdown_grain = breakdown_grain
        self.__token = token
        self.__tables = tables
        self.__save_path = save_path
        self.__username = username
        self.__warehouse = warehouse
        self.__role = role
        self.__database = database
        self.__schema = schema

    # Getter and Setter Methods for Instance Variables
    @property
    def df(self):
        return self.__df

    @df.setter
    def df(self, value):
        self.__df = value

    @property
    def df2(self):
        return self.__df2

    @df2.setter
    def df2(self, value):
        self.__df2 = value

    @property
    def ds1_nm(self):
        return self.__ds1_nm

    @ds1_nm.setter
    def ds1_nm(self, value):
        self.__ds1_nm = value

    @property
    def ds2_nm(self):
        return self.__ds2_nm

    @ds2_nm.setter
    def ds2_nm(self, value):
        self.__ds2_nm = value

    @property
    def case_sens(self):
        return self.__case_sens

    @case_sens.setter
    def case_sens(self, value):
        self.__case_sens = value

    @property
    def print_analysis(self):
        return self.__print_analysis

    @print_analysis.setter
    def print_analysis(self, value):
        self.__print_analysis = value

    @property
    def check_match_by(self):
        return self.__check_match_by

    @check_match_by.setter
    def check_match_by(self, value):
        self.__check_match_by = value

    @property
    def breakdown_grain(self):
        return self.__breakdown_grain

    @breakdown_grain.setter
    def breakdown_grain(self, value):
        self.__breakdown_grain = value

    @property
    def token(self):
        return self.__token

    @token.setter
    def token(self, value):
        self.__token = value

    @property
    def tables(self):
        return self.__tables

    @tables.setter
    def tables(self, value):
        self.__tables = value

    @property
    def save_path(self):
        return self.__save_path

    @save_path.setter
    def save_path(self, value):
        self.__save_path = value

    @property
    def git_org(self):
        return self.__git_org

    @git_org.setter
    def git_org(self, value):
        self.__git_org = value

    def column_gap_analysis(self, df: pd.DataFrame = None, df2: pd.DataFrame = None, ds1_nm: str = 'Source #1',
                            ds2_nm: str = 'Source #2', case_sens: bool = True, print_analysis: bool = True,
                            check_match_by=None, breakdown_grain=None):

        """
        Compares 2 DataFrames and gives shape, size, matching columns, non-matching columns, coverages, and percentages

        :param df: pd.DataFrame, optional, default = None
        :param df2: pd.DataFrame, optional, default = None
        :param ds1_nm: str, optional, default = 'Source #1'
        :param ds2_nm: str, optional, default = 'Source #2'
        :param case_sens: bool, optional, default = True
        :param print_analysis: bool, optional, default = True
        :param check_match_by: any, opioonal, default = None
        :param breakdown_grain: any, optional, default
        :return DataFrame
        """

        if df is None and self.__df:
            df = self.__df
        if df2 is None and self.__df2:
            df2 = self.__df2
        if df2 is None:
            raise ValueError(f'Please insert pandas.DataFrame for df2')
        if ds1_nm == 'Source #1':
            ds1_nm = self.__ds1_nm
        if ds2_nm == 'Source #2':
            ds2_nm = self.__ds2_nm
        if check_match_by is None and self.__check_match_by:
            check_match_by = self.__check_match_by
        if breakdown_grain is None and self.__breakdown_grain:
            breakdown_grain = self.__breakdown_grain

        return _QA.QA.column_gap_analysis(self, df2=df2, ds1_nm=ds1_nm, ds2_nm=ds2_nm, case_sens=case_sens,
                                          print_analysis=print_analysis, check_match_by=check_match_by,
                                          breakdown_grain=breakdown_grain, df1=df)

    def data_prfl_analysis(self, df: pd.DataFrame = None, ds_name: str = 'Data Source', sample_vals: int = 5,
                           print_analysis: bool = True, show_pct_fmt: bool = True):

        """

        :param df: pandas.DataFrame to be analyzed
        :param ds_name: name of the data source to be included in the output
        :param sample_vals:
        :param print_analysis:
        :param show_pct_fmt:
        :return: pandas.Dataframe with the following columns ['DATA_SOURCE', 'COLUMN', 'COL_DATA_TYPE', 'TOTAL_ROWS',
        'ROW_DTYPE_CT', 'PRIMARY_DTYPE_PCT', 'COVERAGE_PCT', 'NULL_PCT', 'DTYPE_ERROR_FLAG', 'NON_NULL_ROWS',
        'NULL_VALUES', 'UNIQUE_VALUES', 'COL_VALUE_SAMPLE', 'NULL_VALUE_SAMPLE']
        """

        if df is None:
            df = self.__df

        return _QA.QA.data_prfl_analysis(self, df=df, ds_name=ds_name, sample_vals=sample_vals,
                                         print_analysis=print_analysis, show_pct_fmt=show_pct_fmt)

    def github_dependencies(self, tables: list | str = None, token: str | None = None, save_path: str = None,
                            org: str = ''):

        if token is None and self.__token is not None:
            token = self.__token
        if tables is None and self.__tables is not None:
            tables = self.__tables
        if org is None and self.__git_org is not None:
            org = self.__git_org

        return _GitHub.GitHub.github_dependencies(self, token=token, tables=tables, save_path=save_path, org=org)

    def search_dependencies(self, tables: list | str = None,
                            token: str | None = None,
                            save_path: str = None,
                            username: str | None = None,
                            warehouse: str | None = None,
                            role: str | None = None,
                            database: str | None = None,
                            schema: str | list | None = None,
                            github_save_path: str | None = None,
                            snowflake_save_path: str | None = None):

        from NikeCA import Snowflake
        import json

        """

        :param token: 
        :param save_path:
        :param tables:
        :param username:
        :param warehouse:
        :param role:
        :param database:
        :param schema:
        :return:
        """

        if tables is None:
            tables = self.__tables
        if username is None:
            username = self.__username
        if warehouse is None:
            warehouse = self.__warehouse
        if role is None:
            role = self.__role
        if database is None:
            database = self.__database
        if schema is None:
            schema = self.__schema
        if save_path is None:
            save_path = self.__save_path
        if token is None:
            token = self.__token
        if tables is None:
            tables = self.__tables

        gh = _GitHub.GitHub.github_dependencies(self, token=token, tables=tables, save_path=github_save_path)
        sf = Snowflake(username=username, warehouse=warehouse, role=role).snowflake_dependencies(tables=tables,
                                                                                                 database=database,
                                                                                                 schema=schema,
                                                                                                 save_path=snowflake_save_path)

        d = {'github_dependencies': gh, 'snowflake_dependencies': sf}

        if save_path is not None:
            with open(save_path, "w+") as f:
                json.dump(d, f, indent=4)

        return d





class BuildSearchQuery:
    pass


def build_search_query(inp_db=None, schema=None, table=None, column_name=None, like_flag=False,
                       col_and_or='AND'):
    """
    Constructs a SQL query for searching tables and columns in a database based on specified search criteria.

    Parameters:
    -----------
    inp_db: str, optional
        The database name to search in. If not specified, searches all databases.
    schema: str, optional
        The schema name to search in. If not specified, searches all schemas.
    table: str, optional
        The table name to search for. If not specified, searches all tables.
    column_name: str or list of str, optional
        The column name(s) to search for. If not specified, searches all columns.
        If a list is provided, searches for any columns that match any of the names in the list.
    like_flag: bool, optional
        If True, uses a SQL LIKE statement to search for columns that contain the specified column name(s).
        If False, searches for exact matches to the specified column name(s).
        If not specified, defaults to False.
    col_and_or: str, optional
        If specified and column_name is a list, determines whether to search for columns that match all or any of
        the names in the list. Must be one of the following values: 'AND', 'and', 'OR', 'or'.
        If not specified, defaults to 'AND'.

    Returns:
    --------
    str
        The constructed SQL query string.
    """

    # ie. always TRUE --> allows us to search for tables/cols/etc. even without knowing the db
    where_stmt = "WHERE 1=1 "
    where_stmt = where_stmt + f"AND table_catalog = '{inp_db}' " if inp_db else where_stmt
    where_stmt = where_stmt + f"AND table_schema = '{schema}' " if schema else where_stmt
    where_stmt = where_stmt + f"AND table_name = '{table}' " if table else where_stmt

    # add column(s) search criteria -->
    # if like_flag = false then column name equals
    # if column_name is a list add an AND statement for each search value in the list
    if column_name is not None:
        if type(column_name) == str:
            where_stmt = where_stmt + f"AND column_name like '{column_name}' " \
                if like_flag else where_stmt + f"AND column_name = '{column_name}' "

        # OR statement where value matches multiple
        elif (type(column_name) == list) & (not like_flag):
            where_stmt = where_stmt + f"""AND column_name in ({' ,'.join(f"'{str(x)}'" for x in column_name)})"""

        # --> user input list of search criteria
        elif type(column_name) == list:
            for idx, x in enumerate(column_name):

                # col contains all column_name search criteria
                if col_and_or.lower() == 'and':
                    where_stmt = where_stmt + f"AND column_name like '{x}' " \
                        if like_flag else where_stmt + f"AND column_name = '{x}' "

                # col contains any of the column_name search criteria
                elif col_and_or.lower() == 'or':
                    where_stmt = where_stmt + f"AND (column_name like '{x}' " \
                        if idx == 0 else where_stmt + f"OR column_name like '{x}' "

                # non-matching input value
                else:
                    raise ValueError('col_and_or input must match: AND/And/and, OR/Or/or')
            where_stmt = where_stmt + ')' if (type(column_name) == list) & (col_and_or == 'or') else where_stmt

        # --> invalid format
        else:
            raise ValueError(f'ERROR: column_name={column_name} does not match required input of list/string')

    # final search-schema query
    query = f'''        
    SELECT 
        DISTINCT
        TABLE_CATALOG
        ,TABLE_SCHEMA
        ,TABLE_NAME
        ,COLUMN_NAME
        ,IS_NULLABLE
        ,DATA_TYPE
    FROM 
        INFORMATION_SCHEMA.COLUMNS
    {where_stmt}
    ORDER BY 
        TABLE_CATALOG
        , TABLE_SCHEMA
        , TABLE_NAME
        , COLUMN_NAME
    '''
    return query







import _BuildSearchQuery
import _SnowflakePull
import _SnowflakeDependencies
import _IMPSummaryDashboard


class SnowflakeData(_IMPSummaryDashboard.IMPSummaryDashboard
                    , _SnowflakePull.SnowflakePull
                    , _SnowflakeDependencies.SnowflakeDependencies
                    , _BuildSearchQuery.BuildSearchQuery):

    import pandas

    def snowflake_pull(self, query: str | dict | None, un, wh, db, role, schema=None, table=None,
                       sample_table: bool = False, sample_val: bool = False, table_sample: dict = None,
                       dtypes_conv=None, separate_dataframes: bool = True): # -> pandas.DataFrame:

        return _SnowflakePull.SnowflakePull.snowflake_pull(self, query=query, un=un, wh=wh, db=db, role=role,
                                                           schema=schema, table=table, sample_table=sample_table,
                                                           sample_val=sample_val, table_sample=table_sample,
                                                           dtypes_conv=dtypes_conv,
                                                           separate_dataframes=separate_dataframes)

    def search_schema(self, un, wh, db, role, sample_table: bool = False, sample_val: bool = False,
                      table_sample: dict = None, dtypes_conv=None, schema=None, table=None, column_name=None,
                      col_and_or='and', get_ex_val=False, like_flag=True):

        import pandas as pd

        # --> pull data, filter out exclusions
        results = pd.DataFrame()
        if type(db) == list:
            queries = {}

            for k in db:
                queries[k] = _BuildSearchQuery.build_search_query(inp_db=k, schema=schema, table=table,
                                                                  column_name=column_name, like_flag=like_flag,
                                                                  col_and_or=col_and_or)
                queries[k] = queries[k][:queries[k].find("INFORMATION_SCHEMA.COLUMNS")] + k + '.' + \
                             queries[k][queries[k].find("INFORMATION_SCHEMA.COLUMNS"):]

            df = _SnowflakePull.SnowflakePull.snowflake_pull(self, queries, un=un, wh=wh, role=role, db=None,
                                                             sample_table=sample_table, sample_val=sample_val,
                                                             table_sample=table_sample, dtypes_conv=dtypes_conv,
                                                             separate_dataframes=False)

            results = pd.concat([results, df], axis=0)

        elif db is None:
            # --> check user's database access for list of dbs to check
            get_dbs = _SnowflakePull.SnowflakePull.snowflake_pull(self, query='''SHOW DATABASES''', un=un, db=db, wh=wh,
                                                                  role=role)

            # list of user's db names
            db_names = list(get_dbs['name'].values)

            print(f"No input database --> checking all of databases in user's access: {len(db_names)} total databases")
            queries = {}
            for db in db_names:
                queries[db] = _BuildSearchQuery.build_search_query(inp_db=db, schema=schema, table=table,
                                                                   column_name=column_name, like_flag=like_flag,
                                                                   col_and_or=col_and_or)
                queries[db] = queries[db][:queries[db].find("INFORMATION_SCHEMA.COLUMNS")] + db + '.' + \
                              queries[db][queries[db].find("INFORMATION_SCHEMA.COLUMNS"):]

            temp_results = _SnowflakePull.SnowflakePull.snowflake_pull(self, query=queries, un=un, wh=wh, role=role,
                                                                       db=None,
                                                                       sample_table=sample_table, sample_val=sample_val,
                                                                       table_sample=table_sample,
                                                                       dtypes_conv=dtypes_conv,
                                                                       separate_dataframes=False)

            results = pd.concat([results, temp_results], axis=0)
        else:
            results = _SnowflakePull.SnowflakePull.snowflake_pull(
                self,
                query=_BuildSearchQuery.build_search_query(inp_db=db, schema=schema, table=table,
                                                           column_name=column_name, like_flag=like_flag,
                                                           col_and_or=col_and_or),
                un=un, db=db, wh=wh, role=role)

        # exclude from table results
        exclusions = ['TEST', 'BACKUP', 'BKUP', 'BCKUP', 'BCKP', '_OLD', 'UPDT', 'DELETED', 'FIX']
        # # drop exclusion rows
        results_fin = results  # [~results['TABLE_NAME'].str.contains('|'.join(exclusions))].copy().reset_index(drop=True)

        # --> print result statement
        print(f'''
            Total table-columns found: {len(results_fin)}    

            Unique column names found: {list(results_fin['COLUMN_NAME'].unique())}
            Total = {len(list(results_fin['COLUMN_NAME'].unique()))}
                ''')

        # --> flagged to retrieve a sample value for each column
        if get_ex_val:
            results_fin['EX_VALS'] = None
            # --> loop through each row & retrieve values
            for indx, row in results_fin.iterrows():
                try:
                    row_res = _SnowflakePull.SnowflakePull.snowflake_pull(self, '', un=un, db=None, wh=wh, role=role,
                                                                          sample_val=True,
                                                                          table_sample={'db': row['TABLE_CATALOG'],
                                                                                        'schema': row['TABLE_SCHEMA'],
                                                                                        'table': row['TABLE_NAME'],
                                                                                        'col': row[
                                                                                            'COLUMN_NAME']})  # row results
                    # set row example values equal to unique column value list
                    row['EX_VALS'] = list(row_res[row['COLUMN_NAME']].unique())
                except:
                    print(f"Could not pull {row['COLUMN_NAME']} for table: {row['TABLE_NAME']}")
                    continue

        return results_fin

    def snowflake_dependencies(self, tables: str | list, username: str, warehouse: str, role: str,
                               database: str | None = None, schema: str | list | None = None, save_path: str = None):

        return _SnowflakeDependencies.SnowflakeDependencies.snowflake_dependencies(self, tables=tables,
                                                                                   username=username,
                                                                                   warehouse=warehouse, role=role,
                                                                                   database=database, schema=schema,
                                                                                   save_path=save_path)

    def optimize_tbl_mem(self, username: str, warehouse: str, role: str, database: str = None, schema: str = None,
                         table_name: str = None, pull_all_cols=True, run_debugging: bool = False,
                         query=None):

        import numpy as np
        from NikeQA import QA

        from datetime import datetime

        if query is not None:
            query = str(query) + ' LIMIT 5' if query[-1] != ';' else str(
                query[:-1]) + ' LIMIT 5'  # add limit if user inputs a query (checking for semicolon at end of query)
            print(query)

        # --> DEBUGGING CODE: Return data profile if run for debugging purposes
        if run_debugging:
            t_sample = self.snowflake_pull(query=None, un=username, wh=warehouse, role=role, db=database,
                                           sample_table=True,
                                           table_sample={'db': database, 'schema': schema, 'table': table_name}
                                           if query is None else query)

            sample_prfl = QA(t_sample).data_prfl_analysis(ds_name='Table Sample', print_analysis=False)  # --> run data profiling function to get table info
            return sample_prfl

        # ======================================================================================================
        # --> DYNAMICALLY DETERMINE HOW TO CONVERT COLUMN DATA TYPES TO OPTIMIZE PYTHON-PANDAS MEMORY
        t_sample = self.snowflake_pull(query=None, un=username, wh=warehouse, role=role, db=database, sample_table=True,
                                       table_sample={'db': database, 'schema': schema, 'table': table_name}) \
            if query is None else self.snowflake_pull(query, username, warehouse, database, role=role)  # --> input table information from function inputs & sample table

        # --> determine column datatype conversions to make:
        sample_prfl = QA(t_sample).data_prfl_analysis(ds_name='Table Sample', print_analysis=False)  # --> run data profiling function to get table info

        sample_prfl['UNIQUE_PCT'] = (sample_prfl['UNIQUE_VALUES'] / (sample_prfl['NON_NULL_ROWS'])) * 100  # --> determine 0-100 percentage of unique values in the column

        for index, row in sample_prfl.iterrows():  # --> get the min/max values for all our integer columns
            sample_prfl.loc[index, 'INT_MIN'] = min(t_sample[row['COLUMN']]) if row['COL_DATA_TYPE'] == int else None
            sample_prfl.loc[index, 'INT_MAX'] = max(t_sample[row['COLUMN']]) if row['COL_DATA_TYPE'] == int else None

        # drop unneeded columns
        sample_prfl = sample_prfl[['COLUMN', 'COL_DATA_TYPE', 'UNIQUE_PCT', 'INT_MIN', 'INT_MAX']].copy()

        # --> run conversion rules to replace data types
        #    strings/keys/anything <> int w/ a distinct record on each row (>7% for our logic) = pd.StringDtype()
        #    integers = int64 (error handling), except 'int8' = 'int8'
        #    any float or decimals = float32
        #    any string/object that is not distinct on each row = 'category'
        #    any True/False fields = bool ('category' may also work)
        sample_prfl['DTYPE_CONV'] = np.where(
            ((sample_prfl['COL_DATA_TYPE'] == object) & (sample_prfl['UNIQUE_PCT'] <= 66)) | (
                        (sample_prfl['COL_DATA_TYPE'] == object) & (sample_prfl['UNIQUE_PCT'].isna())), 'category',
            # default string/object value
            np.where((sample_prfl['COL_DATA_TYPE'] == object) & (sample_prfl['UNIQUE_PCT'] > 66), str, #, pd.StringDtype(),
                     # objects w/ a distinct record on each row (>80% for our logic) = pd.StringDtype()
                     np.where(sample_prfl['COL_DATA_TYPE'] == int, 'int64',
                              # handles dtype error (makes other INT criteria irrelevant)
                              np.where(sample_prfl['COL_DATA_TYPE'] == 'int8', 'int8',  # default for any integer value
                                       np.where(sample_prfl['COL_DATA_TYPE'] == float, 'float32',  # float columns
                                                np.where(sample_prfl['COL_DATA_TYPE'] == bool, 'bool',
                                                         'ERROR'))))))  # True/False boolean columns

        # --> QA/ERROR checking
        error_flag = sample_prfl.loc[sample_prfl['DTYPE_CONV'] == 'ERROR'].copy().reset_index(drop=True)
        if len(error_flag) > 0:
            raise ValueError(f'''ERROR: the following columns have no data type conversion rule: {
                list(error_flag['COLUMN'].unique())}''')  # raise error if we're missing any conversion rules

        sample_prfl.index = sample_prfl['COLUMN']  # reset index to column name
        sample_prfl.drop(columns=['COLUMN', 'COL_DATA_TYPE', 'UNIQUE_PCT', 'INT_MIN', 'INT_MAX'],
                         inplace=True)  # drop unneeded columns

        dtypes = sample_prfl.to_dict()['DTYPE_CONV']  # convert to dictionary --> use for final query

        # ======================================================================================================
        # --> ANALYZE MEMORY OPTIMIZATION: get sample after memory conversion
        t_test = self.snowflake_pull(None, un=username, wh=warehouse, role=role, db=database, sample_table=True,
                                     table_sample={'db': database, 'schema': schema, 'table': table_name},
                                     dtypes_conv=dtypes) if query is None else self.snowflake_pull(
            query=query, un=username, wh=warehouse, role=role, db=database, dtypes_conv=dtypes)

        before_memory = t_sample.memory_usage(deep=True).sum()  # sample table memory usage (before conversion)
        after_memory = t_test.memory_usage(deep=True).sum()  # sample table memory usage (after conversion)

        print(f'''
        Sample table memory usage -->
        Before memory conversion: {before_memory / 1000000000} GB
        After memory conversion: {after_memory / 1000000000} GB
        Memory reduction percentage: {"{:.2%}".format((before_memory - after_memory) / before_memory)}

        Finished running! {datetime.now().strftime("%H:%M:%S")}
        ''')  # convert to GB & print total memory usage before & after conversion

        return dtypes  # --> return the datatypes to convert each column in the table to


